from datalog.Term import Variable, Constant, FunctionalTerm
from datalog.Predicate import Predicate
from datalog.Rule import Rule
from abc import ABC, abstractmethod

class SQLStrategy(ABC):
    
    def __init__(self, program, srcManager):
        self._srcManager = srcManager
        self._program = program
        self._newPlaceHolder = "$new$"
        self._deltaPlaceHolder = "$delta$"
        self._oldPlaceHolder = "$old$"
        self._targetPredicate = None #Predicate: target predicate
        self._targetColumns = [] #Column names of target table
        self._termsToSelect = [] #list: terms to select (i.e., the head schema)
        self._bodyPreds = [] #list: ordered list of body predicates 
        self._recursiveBodyPreds = [] #list: {boolean: true if predicate i is recursive}
        self._nRecursiveBodyPreds = 0
        self._varsFrom = {} #dict: (key=variable,value=(body_lit_pos,term_pos))
        self._bodyTermPos = {} #dict: (key=term,value=position_list)
        self._predSchema = {} #dict: (key=(predName,predArity),value=attribute list)
        self._recursiveRuleCache = {} #dict: (key=ruleIdx, value=SQLEquivalentQuery)

    def rewriteExitRule(self,rule):
        assert isinstance(rule, Rule)
        assert len(rule.getHead()) == 1 
        for atom in rule.getHead():
            self.__addToHead(atom)
        for lit in rule.getBody():
            self.__addToBody(lit)
        sql = self._getExitRuleInSQL()
        self.__reset()
        return sql
    
    def rewriteRecursiveRule(self,ruleIdx,rule,recursionLevel,program):
        assert isinstance(rule, Rule)
        assert len(rule.getHead()) == 1
        sql = ""
        if ruleIdx in self._recursiveRuleCache:
            sql = self._recursiveRuleCache[ruleIdx]
        else:
            for atom in rule.getHead():
                self.__addToHead(atom)
            for lit in rule.getBody():
                self.__addToBody(lit)
                if program.getComponent(lit.getPredicate()) == program.getComponent(rule.getHead()[0].getPredicate()):
                    self._recursiveBodyPreds.append(True)
                    self._nRecursiveBodyPreds += 1
                else:
                    self._recursiveBodyPreds.append(False)
            sql = self._getRecursiveRuleInSQL()
            self._recursiveRuleCache[ruleIdx] = sql
            self.__reset()

        sql1 = sql.replace(self._newPlaceHolder, str(recursionLevel + 1))
        sql2 = sql1.replace(self._deltaPlaceHolder, str(recursionLevel))
        sql3 = sql2.replace(self._oldPlaceHolder, str(recursionLevel - 1))
        return sql3

    def startRecursiveSubprogram(self):
        print("\nStart recursive subProgram.")
        pass
    
    def endRecursiveSubprogram(self):
        self._recursiveRuleCache.clear()

    @abstractmethod
    def _getExitRuleInSQL(self):
        pass

    @abstractmethod
    def getNumberRowsInserted(self, parameter1, parameter2=None):
        pass
    
    @abstractmethod
    def _getRecursiveRuleInSQL(self):
        pass

    @abstractmethod
    def getTuples(self,predicate):
        pass

    @abstractmethod
    def dropTable(self, predicate):
        pass

    def _getTableName(self,pred):
        name = pred.getName()
        arity = pred.getArity()
        if (name,arity) not in self._predSchema:
            self._predSchema[(name,arity)] = self._srcManager.getTableSchema(pred)
        return name
    
    def _getTableAlias(self,srcIndex):
        return "t"+str(srcIndex)

    def _getAttrName(self,bodyLitPos,termPos):
        assert bodyLitPos in range(len(self._bodyPreds))
        assert termPos in range(self._bodyPreds[bodyLitPos].getArity())
        predName = self._bodyPreds[bodyLitPos].getName()
        predArity = self._bodyPreds[bodyLitPos].getArity()  
        if (predName,predArity) not in self._predSchema:
            self._predSchema[(predName,predArity)] = self._srcManager.getTableSchema(self._bodyPreds[bodyLitPos])
        return self._predSchema[(predName,predArity)][termPos]
    
    def __addToHead(self,atom):
        self._targetPredicate = atom.getPredicate()
        self._termsToSelect = atom.getTerms()
        if (self._targetPredicate.getName(),self._targetPredicate.getArity()) not in self._predSchema:
            self._predSchema[(self._targetPredicate.getName(),self._targetPredicate.getArity())] = self._srcManager.getTableSchema(self._targetPredicate)
        self._targetColumns = self._predSchema[(self._targetPredicate.getName(),self._targetPredicate.getArity())]

    def __addToBody(self,lit):
        self._bodyPreds.append(lit.getPredicate())
        litPos = len(self._bodyPreds)-1
        for i in range(len(lit.getTerms())):
            t = lit.getTerms()[i]
            #first check whether t is a variable occurring in the head
            if isinstance(t,Variable):
                if t in self._termsToSelect:
                    if t not in self._varsFrom: #if not already seen
                        self._varsFrom[t] = (litPos,i)
                        # memorizzare in che posizione
            #now handle t as a body term
            if t not in self._bodyTermPos:
                self._bodyTermPos[t] = [(litPos,i)]
            else:
                self._bodyTermPos[t].append((litPos,i))

    def __reset(self):
        self._targetPredicate = None
        self._termsToSelect.clear()
        self._bodyPreds.clear()
        self._varsFrom.clear()
        self._bodyTermPos.clear()
        self._recursiveBodyPreds.clear()
        self._nRecursiveBodyPreds = 0
        self._targetColumns = []